/*
 *	HT Editor
 *	htobj.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htcurses.h"
#include "htdebug.h"
#include "hthist.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htmeminf.h"
#include "htobj.h"
#include "htpal.h"
#include "htsearch.h"
#include "htstring.h"
#include "httag.h"
#include "tools.h"
#include "x86dis.h"

#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#define zerofillchar '.'

/*
 *	CLASS ht_view
 */

ht_view::ht_view(bounds *b, int _options, char *_desc)
: object()
{
	desc=strdup(_desc);
	group=0;
	focused=0;
	browse_idx=0;
	view_is_dirty=1;
	size.x=0;
	size.y=0;
	size.w=0;
	size.h=0;
	prev=0;
	next=0;
	setoptions(_options);
	buf=0;
	enabled=1;
	if (options & hto_ownbuffer) {
		buf=new drawbuf(&size);
		enable_buffering();
	} else {
		buf=screen;
		disable_buffering();
	}
	setbounds(b);
	palette();
}

ht_view::~ht_view()
{
	if (desc) delete desc;
}

int ht_view::childcount()
{
	return 1;
}

void ht_view::cleanview()
{
	view_is_dirty=0;
}

void ht_view::clear(char color)
{
	buf->fill(size.x, size.y, size.w, size.h, color, ' ');
}

void ht_view::clearmsg(htmsg *msg)
{
	msg->msg=htm_nothing;
	msg->type=htmt_nothing;
}

int ht_view::countselectables()
{
	return (options&hto_selectable)!=0;
}

int ht_view::datasize()
{
	return 0;
}

void ht_view::dirtyview()
{
	view_is_dirty=1;
}

void ht_view::disable()
{
	enabled=0;
}

void ht_view::disable_buffering()
{
	if (options & hto_ownbuffer) {
		if (buf) delete buf;
		buf=screen;
		setoptions(options&(~hto_ownbuffer));
	}
}

STUB void ht_view::draw()
{
}

void ht_view::enable()
{
	enabled=1;
}

void ht_view::enable_buffering()
{
	if (!(options & hto_ownbuffer)) {
		buf=new drawbuf(&size);
		setoptions(options|hto_ownbuffer);
	}
}

void ht_view::fill(int x, int y, int w, int h, int c, int chr)
{
	buf->fill(size.x+x, size.y+y, w, h, c, chr);
}

int ht_view::focus(ht_view *view)
{
	if (view==this) {
		if (!focused) receivefocus();
		return 1;
	}
	return 0;
}

ABSTRACT void ht_view::getdata(void *buf)
{
	abstract();
}

void ht_view::getbounds(bounds *b)
{
	*b=size;
}

ht_view *ht_view::getselected()
{
	return this;
}

STUB ht_view *ht_view::getfirstchild()
{
	return 0;
}

int ht_view::getnumber()
{
	return -1;
}

void ht_view::handlemsg(htmsg *msg)
{
	switch (msg->msg) {
		case htm_draw:
			redraw();
			break;
		case htm_dirtyview:
			dirtyview();
			if (msg->type & htmt_broadcast==0) clearmsg(msg);
			break;
	}
}

void ht_view::hidecursor()
{
	screen->hidecursor();
}

int ht_view::alone()
{
	return (group && group->isalone(this));
}

void ht_view::autopalette()
{
}

int ht_view::isalone(ht_view *view)
{
	return (view==this) && countselectables();
}

int ht_view::isviewdirty()
{
	return view_is_dirty;
}

int ht_view::lprint(int x, int y, int c, int l, char *text)
{
	return buf->lprint(size.x+x, size.y+y, c, l, text);
}

int ht_view::lprintw(int x, int y, int c, int l, int *text)
{
	return buf->lprintw(size.x+x, size.y+y, c, l, text);
}

void ht_view::move(int rx, int ry)
{
	size.x+=rx;
	size.y+=ry;
	buf->rmove(rx, ry);
}

int ht_view::printf(int x, int y, int c, char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return print(x, y, c, buf);
}

int ht_view::print(int x, int y, int c, char *text)
{
	return buf->print(size.x+x, size.y+y, c, text);
}

int ht_view::printw(int x, int y, int c, int *text)
{
	return buf->printw(size.x+x, size.y+y, c, text);
}

void ht_view::printchar(int x, int y, int c, int ch)
{
	buf->printchar(size.x+x, size.y+y, c, ch);
}

void ht_view::receivefocus()
{
	dirtyview();
	focused=1;
}

void ht_view::redraw()
{
	if (options & hto_ownbuffer) {
		if (isviewdirty()) {
			draw();
			cleanview();
		}
		screen->drawbuffer((drawbuf*)buf, size.x, size.y);
	} else {
		draw();
		cleanview();
	}
}

void ht_view::releasefocus()
{
	dirtyview();
	hidecursor();
	focused=0;
}

int ht_view::select(ht_view *view)
{
	return (view==this);
}

void ht_view::selectfirst()
{
}

void ht_view::selectlast()
{
}

void ht_view::sendmsg(htmsg *msg)
{
	handlemsg(msg);
}

void ht_view::sendmsg(int msg, int data)
{
	htmsg m;
	switch (msg) {
		case htm_nothing:
			return;
		case htm_draw:
		case htm_dirtyview:
			m.msg=msg;
			m.type=htmt_broadcast;
			m.data_int=data;
			break;
		default:
			m.type=htmt_nothing;
			m.msg=msg;
			m.data_int=data;
			break;
	}
	sendmsg(&m);
}

void ht_view::setbounds(bounds *b)
{
	size=*b;
	if (options & hto_ownbuffer) {
		buf->setbounds(&size);
	}
}

void ht_view::setcursor(int x, int y)
{
	screen->setcursor(size.x+x, size.y+y);
}

ABSTRACT void ht_view::setdata(void *buf)
{
	abstract();
}

void ht_view::setoptions(int _options)
{
	options=_options;
}

void ht_view::setpalette(palette *_pal)
{
	pal=_pal;
}

/*
 *	CLASS ht_group
 */

ht_group::ht_group(bounds *b, int options, char *desc)
: ht_view(b, options, desc)
{
	first=0;
	current=0;
	last=0;
	view_count=0;
	shared_data=0;
/* FIXME: debug */
	if (options & hto_ownbuffer) {
		WARN("class 'ht_group' should not have own buffer !");
	}
/**/
}

ht_group::~ht_group()
{
	ht_view *a, *b;
	a=first;
	while (a) {
		b=a->next;
		delete a;
		a=b;
	}
}

int ht_group::childcount()
{
	return view_count;
}

int ht_group::countselectables()
{
	int c=0;
	ht_view *v=first;
	while (v) {
		c+=v->countselectables();
		v=v->next;
	}
	return c;
}

STUB int ht_group::focus(ht_view *view)
{
	ht_view *v=first;
	while (v) {
		if (v->focus(view)) {
			releasefocus();
			current=v;
			putontop(v);
			receivefocus();
			return 1;
		}
		v=v->next;
	}
	return ht_view::focus(view);
}

int ht_group::focusnext()
{
	int i=current->browse_idx;
	if ((i==view_count-1) && !alone()) {
		return 0;
	}
	while (1) {
		i++;
		if (i>view_count-1) i=0;
		if (i==current->browse_idx) break;
		ht_view *v=get_by_browse_idx(i);
		if (v->options & hto_selectable) {
			v->selectfirst();
			focus(v);
			return 1;
		}
	}
	return 0;
}

int ht_group::focusprev()
{
	int i=current->browse_idx;
	if (!i && !alone()) {
		return 0;
	}
	while (1) {
		i--;
		if (i<0) i=view_count-1;
		if (i==current->browse_idx) break;
		ht_view *v=get_by_browse_idx(i);
		if (v->options & hto_selectable) {
			v->selectlast();
			focus(v);
			return 1;
		}
	}
	return 0;
}

ht_view *ht_group::get_by_browse_idx(int i)
{
	ht_view *v=first;
	while (v) {
		if (v->browse_idx==i) return v;
		v=v->next;
	}
	return 0;
}

ht_view *ht_group::getselected()
{
	return current->getselected();
}

ht_view *ht_group::getfirstchild()
{
	return first;
}

STUB void ht_group::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_broadcast) {
		ht_view::handlemsg(msg);
		ht_view *v=first;
		while (v) {
			v->handlemsg(msg);
			v=v->next;
		}
	} else {
		int msgtype=msg->type;
		ht_view *v;

		msg->type=htmt_preprocess;
		v=first;
		while (v) {
			if (v->options & hto_preprocess) {
				v->handlemsg(msg);
			}
			v=v->next;
		}

		msg->type=htmt_nothing;
		current->handlemsg(msg);

		msg->type=htmt_postprocess;
		v=first;
		while (v) {
			if (v->options & hto_postprocess) {
				v->handlemsg(msg);
			}
			v=v->next;
		}

		msg->type=msgtype;
		ht_view::handlemsg(msg);
	}
	if (((msg->type==htmt_nothing) || (msg->type==htmt_broadcast)) && (msg->msg==htm_keypressed)) {
		switch (msg->data_int) {
			case K_BackTab: {
				if (focusprev()) {
					clearmsg(msg);
					dirtyview();
				}
				break;
			}
			case K_Tab: {
				if (focusnext()) {
					clearmsg(msg);
					dirtyview();
				}
				break;
			}
		}
	}
}

void ht_group::move(int rx, int ry)
{
	ht_view::move(rx, ry);
	ht_view *v=first;
	while (v) {
		v->move(rx, ry);
		v=v->next;
	}
}

void ht_group::insert(ht_view *view)
{
	if (view->options & hto_preprocess) setoptions(options | hto_preprocess);
	if (view->options & hto_postprocess) setoptions(options | hto_postprocess);
	view->move(size.x, size.y);
	if (last) last->next=view;
	view->prev=last;
	view->next=0;
	last=view;
	if (!first) first=view;
	view->group=this;
	view->browse_idx=view_count++;
	if ((!current) || ((current) && (!(current->options & hto_selectable)) && (view->options & hto_selectable))) {
		current=view;
	}
	if ((current) && (current->options & hto_selectable)) {
		if (focused) {
			focus(current);
		} else {
			select(current);
		}
	}
}

int ht_group::isalone(ht_view *view)
{
	ht_view *v=first;
	while (v) {
		if ((v!=view) && (v->countselectables())) return 0;
		v=v->next;
	}
	return 1;
}

int ht_group::isviewdirty()
{
	ht_view *v=first;
	while (v) {
		if (v->isviewdirty()) return 1;
		v=v->next;
	}
	return 0;
}

void ht_group::putontop(ht_view *view)
{
	if (view->next) {
		if (view->prev) view->prev->next=view->next; else first=view->next;
		view->next->prev=view->prev;
		view->prev=last;
		view->next=0;
		last->next=view;
		last=view;
	}
}

void ht_group::receivefocus()
{
	ht_view::receivefocus();
	current->receivefocus();
}

STUB void ht_group::redraw()
{
	ht_view::redraw();
	ht_view *v=first;
	while (v) {
		v->redraw();
		v=v->next;
	}
}

void ht_group::releasefocus()
{
	ht_view::releasefocus();
	current->releasefocus();
}

void ht_group::remove(ht_view *view)
{
	if (view->prev) view->prev->next=view->next;
	if (view->next) view->next->prev=view->prev;
	if (first==view) first=first->next;
	if (last==view) last=last->prev;
}

STUB int ht_group::select(ht_view *view)
{
	ht_view *v=first;
	while (v) {
		if (v->select(view)) {
			current=v;
			putontop(v);
			return 1;
		}
		v=v->next;
	}
	return ht_view::select(view);
}

void ht_group::selectfirst()
{
	for (int i=0; i<view_count; i++) {
		ht_view *v=first;
		while (v) {
			if ((v->browse_idx==i) && (v->options&hto_selectable)) {
				select(v);
				return;
			}
			v=v->next;
		}
	}
}

void ht_group::selectlast()
{
	for (int i=view_count-1; i>=0; i--) {
		ht_view *v=first;
		while (v) {
			if ((v->browse_idx==i) && (v->options&hto_selectable)) {
				select(v);
				return;
			}
			v=v->next;
		}
	}
}

void ht_group::setpalette(palette *_pal)
{
	ht_view *v=first;
	while (v) {
		v->setpalette(_pal);
		v=v->next;
	}
	ht_view::setpalette(_pal);
}

/*
 *	CLASS ht_xgroup
 */

ht_xgroup::ht_xgroup(bounds *b, int options, char *desc)
: ht_group(b, options, desc)
{
	first=0;
	current=0;
	last=0;
}

ht_xgroup::~ht_xgroup()
{
}

int ht_xgroup::countselectables()
{
	return current->countselectables();
}

STUB void ht_xgroup::handlemsg(htmsg *msg)
{
/* no broadcasts. */
	if (current) current->handlemsg(msg);
	ht_view::handlemsg(msg);
}

int ht_xgroup::isalone(ht_view *view)
{
	if (group) return group->isalone(this);
	return 0;
}

void ht_xgroup::redraw()
{
	ht_view::redraw();
/* no broadcasts. */
	if (current) current->redraw();
}

void ht_xgroup::selectfirst()
{
	current->selectfirst();
}

void ht_xgroup::selectlast()
{
	current->selectlast();
}

/*
 *	CLASS ht_scrollbar
 */

ht_scrollbar::ht_scrollbar(bounds *b, char *_color)
: ht_view(b, hto_ownbuffer, "some scrollbar")
{
	color=_color;
	ppos=0;
}

ht_scrollbar::~ht_scrollbar()
{
}

void ht_scrollbar::enable()
{
	enable_buffering();
	ht_view::enable();
	dirtyview();
}

void ht_scrollbar::disable()
{
	disable_buffering();
	ht_view::disable();
	dirtyview();
}

void ht_scrollbar::draw()
{
	if (enabled) {
		printchar(0, 0, *color, CHAR_ARROWBIG_UP);
		buf->fill(size.x, size.y+1, size.w, size.h-2, *color, CHAR_FILLED_M);
		printchar(0, size.h-1, *color, CHAR_ARROWBIG_DOWN);
		printchar(0, ppos+1, *color, CHAR_QUAD_SMALL);
	}
}

void ht_scrollbar::setpos(int percent)
{
	ppos=(size.h-3)*percent/100;
	dirtyview();
}

/*
 *	CLASS ht_frame
 */

ht_frame::ht_frame(bounds *b, char *desc, int _number)
: ht_view(b, 0, desc)
{
	number=_number;
	setpalette(pal_grp->get(window_palette_str));
}

ht_frame::~ht_frame()
{
}

STUB void ht_frame::draw()
{
	int c=windialog_color_frame;
/* "...Ŀ" */
	printchar(0, 0, c, CHAR_CORNERUL);
	for (int i=1; i<size.w-1; i++) printchar(i, 0, c, CHAR_LINEH);
	printchar(0+size.w-1, 0, c, CHAR_CORNERUR);
/* "..." */
	printchar(0, size.h-1, c, CHAR_CORNERLL);
	for (int i=1; i<size.w-1; i++) printchar(i, size.h-1, c, CHAR_LINEH);
	printchar(size.w-1, size.h-1, c, CHAR_CORNERLR);
/* "", "" */
	for (int i=1; i<size.h-1; i++) {
		printchar(0, i, c, CHAR_LINEV);
		printchar(size.w-1, i, c, CHAR_LINEV);
	}
/* "[x]" */
	print(2, 0, c, "[ ]");
	printchar(3, 0, windialog_color_killer, CHAR_QUAD_SMALL);
/* <title> */
	if (desc) {
		int l=strlen(desc), o=0;
		if (l>size.w-15) {
			o=l-size.w+15;
			l-=o;
			printf((size.w-l-5)/2, 0, c, " ...%s ", desc+o);
		} else {
			printf((size.w-l-2)/2, 0, c, " %s ", desc);
		}
	}
/* e.g. "1" */
	if (number>0) {
		int l=number;
		int i=0;
		do {
			l=l/10;
			i++;
		} while (l);
		printf(size.w-2-i, 0, c, "%d", number);
	}
}

STUB void ht_frame::handlemsg(htmsg *msg)
{
	ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_window
 */

ht_window::ht_window(bounds *b, char *desc, int _number)
: ht_group(b, hto_selectable | hto_browsable, desc)
{
	number=_number;
	scrollbar=0;
	setpalette(pal_grp->get(window_palette_str));
	bounds c=*b;
	c.x=0;
	c.y=0;
	frame=new ht_frame(&c, desc, _number);
	insert(frame);
}

ht_window::~ht_window()
{
}

void ht_window::draw()
{
	clear(windialog_color_body);
}

int ht_window::getnumber()
{
	return number;
}

void ht_window::setscrollbar(ht_scrollbar *_scrollbar)
{
	scrollbar=_scrollbar;
	insert(scrollbar);
	putontop(scrollbar);
}

/*
 *	CLASS ht_winstatictext
 */

ht_winstatictext::ht_winstatictext(bounds *b, char *_text)
: ht_view(b, hto_ownbuffer, "some statictext")
{
	text=strdup(_text);
	setpalette(pal_grp->get(window_palette_str));
}

ht_winstatictext::~ht_winstatictext()
{
	if (text) delete text;
}

void ht_winstatictext::draw()
{
	clear(windialog_color_body);
	char *t=gettext(), *u=t;
	int i=0;
	do {
		u=t;
		t+=lprint(0, i, windialog_color_body, size.w, t);
		i++;
	} while ((*t) && (i<size.h));
}

char *ht_winstatictext::gettext()
{
	return text;
}

void ht_winstatictext::settext(char *_text)
{
	delete text;
	text=strdup(_text);
}

/*
 *	CLASS ht_viewer
 */

ht_viewer::ht_viewer(bounds *b, char *desc, htstream *_file, int _caps)
:ht_view(b, hto_ownbuffer | hto_browsable | hto_selectable, desc)
{
	caps=_caps;
	file=_file;
	shared_data=0;
	setpalette(pal_grp->get(window_palette_str));
}

ht_viewer::~ht_viewer()
{
}

STUB void ht_viewer::complete()
{
}

STUB int ht_viewer::cursor_up(int n)
{
	return 0;
}

STUB int ht_viewer::cursor_down(int n)
{
	return 0;
}

STUB int ht_viewer::cursor_left(int n)
{
	return 0;
}

STUB int ht_viewer::cursor_right(int n)
{
	return 0;
}

STUB void ht_viewer::cursor_home()
{
}

STUB void ht_viewer::cursor_end()
{
}

STUB char *ht_viewer::func(int i, int getname)
{
	return app->func(i, getname);
}

STUB void ht_viewer::handlemsg(htmsg *msg)
{
	int i=0;
	switch (msg->msg) {
		case htm_keypressed: {
			switch (msg->data_int) {
				case K_F12: i++;
				case K_F11: i++;
				case K_F10: i++;
				case K_F9: i++;
				case K_F8: i++;
				case K_F7: i++;
				case K_F6: i++;
				case K_F5: i++;
				case K_F4: i++;
				case K_F3: i++;
				case K_F2: i++;
				case K_F1: {
					i++;
					htmsg m;
					m.msg=htm_funcname;
					m.data_int=i;
					sendmsg(&m);
					if ((m.msg==htm_funcname_return) && (m.data_str)) {
						sendmsg(htm_funcexec, i);
						clearmsg(msg);
					}
					break;
				}
				case K_Up:
					cursor_up(1);
					clearmsg(msg);
					dirtyview();
					break;
				case K_Down:
					cursor_down(1);
					clearmsg(msg);
					dirtyview();
					break;
				case K_Left:
					cursor_left(1);
					clearmsg(msg);
					dirtyview();
					break;
				case K_Right:
					cursor_right(1);
					clearmsg(msg);
					dirtyview();
					break;
				case K_PageUp: {
					bounds b;
					getbounds(&b);
					cursor_up(b.h-1);
					clearmsg(msg);
					dirtyview();
					break;
				}
				case K_PageDown: {
					bounds b;
					getbounds(&b);
					cursor_down(b.h-1);
					clearmsg(msg);
					dirtyview();
					break;
				}
				case K_Control_Left:
					scroll_left(1);
					clearmsg(msg);
					dirtyview();
					break;
				case K_Control_Right:
					scroll_right(1);
					clearmsg(msg);
					dirtyview();
					break;
				case K_Home:
					cursor_home();
					clearmsg(msg);
					dirtyview();
					break;
				case K_End:
					cursor_end();
					clearmsg(msg);
					dirtyview();
					break;
				case K_BackSpace:
					backref();
					clearmsg(msg);
					dirtyview();
					break;
				case K_Space:
					ref();
					clearmsg(msg);
					dirtyview();
					break;
			}
			break;
		}
		case htm_vs_restore:
			view_state_load(msg->data_ptr);
			clearmsg(msg);
			break;
		case htm_funcexec:
			func(msg->data_int, 0);
			clearmsg(msg);
			break;
		case htm_funcname:
			msg->msg=htm_funcname_return;
			msg->data_str=func(msg->data_int, 1);
			break;
		case htm_complete:
			complete();
			break;
	}
	ht_view::handlemsg(msg);
}

STUB int ht_viewer::jump(ADDR ofs, int to_history, ht_view *focused)
{
	return 0;
}

STUB int ht_viewer::jumpstr(char *str)
{
	ADDR v;
	if (hex2int(str, &v)) {
		return jump(v, 1, this);
	}
	return 0;
}

STUB int ht_viewer::ref()
{
	return 0;
}

STUB void ht_viewer::scroll_up(int n)
{
}

STUB void ht_viewer::scroll_down(int n)
{
}

STUB void ht_viewer::scroll_left(int n)
{
}

STUB void ht_viewer::scroll_right(int n)
{
}

void ht_viewer::view_state_delete(void *view_state)
{
	free(view_state);
}

STUB void ht_viewer::view_state_load(void *data)
{
}

STUB void *ht_viewer::view_state_store()
{
	return 0;
}

/*
 *	CLASS ht_uviewer
 */

ht_uviewer::ht_uviewer(bounds *b, char *_desc, htstream *_file, int _caps)
: ht_viewer(b, _desc, _file, _caps)
{
	firstsub=0;
	currentsub=0;
	lastsub=0;
	ofs=0;
	yofs=0;
	xofs=0;
	cursor_sub=0;
	cursor_eofs=0;
	cursor_ofs=0;
	cursor_yofs=0;
	cursor_group=0;
	cursor_e=0;
	cursor_x=0;
	cursor_y=0;
	edit=0;
}

ht_uviewer::~ht_uviewer()
{
	hts_sub *s=firstsub, *t;
	while (s) {
		t=s->next;
		delete s;
		s=t;
	}
}

char *ht_uviewer::func(int i, int getname)
{
	switch (i) {
		case 3:
			if (caps & htvc_edit) {
				if (edit) {
					if (getname) return "view"; else sendmsg(htm_view_mode, 0);
				} else {
					if (getname) return "edit"; else sendmsg(htm_edit_mode, 0);
				}
			}
			return 0;
		case 5:
			if (caps & htvc_goto) {
				if (getname) return "goto"; else jump_dialog(this);
			}
			return 0;
		case 7:
			if (caps & htvc_search) {
				if (getname) return "search"; else search_dialog(this, currentsub, cursor_eofs);
			}
			return 0;
		case 9:
			if (caps & htvc_edit) {
				if (edit) {
					if (getname) return "update"; else update_edit();
				} else {
					if (getname) return "~update";
				}
			}
			return 0;
	}
	return ht_viewer::func(i, getname);
}

void ht_uviewer::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Home:
				cursor_home();
				clearmsg(msg);
				dirtyview();
				break;
			case K_End:
				cursor_end();
				clearmsg(msg);
				dirtyview();
				break;
			case K_Tab:
				cursor_next_group();
				clearmsg(msg);
				dirtyview();
				break;
			case K_Escape:
				if (edit) {
					sendmsg(htm_view_mode, 0);
					clearmsg(msg);
					dirtyview();
				}
				break;
			case K_Control_PageUp: {
				char *e;
				hts_sub *s=firstsub;
				char str[1024];
				ADDR o;
				int y;
				s->start_line_address(&o, &y);
				while (1) {
					s->getline(str, o, y);
					e=get_editable(str, 0, cursor_group);
					if (e) break;
					seek_next_line(&s, &o, &y, 1);
				}
				strcpy(curline, str);
				currentsub=s;
				cursor_sub=s;
				cursor_e=0;
				cursor_group=0;
				cursor_x=0;
				cursor_y=0;
				cursor_ofs=o;
				cursor_yofs=y;
				ofs=o;
				yofs=y;
				update_cursor_data();
				clearmsg(msg);
				dirtyview();
				break;
			}
			case K_Control_PageDown: {
				char *e;
				hts_sub *s=lastsub;
				char str[1024];
				ADDR o;
				int y;
				s->end_line_address(&o, &y);
				while (1) {
					s->getline(str, o, y);
					e=get_editable(str, 0, cursor_group);
					if (e) break;
					seek_prev_line(&s, &o, &y, 1);
				}
				strcpy(curline, str);
				currentsub=s;
				cursor_sub=s;
				cursor_e=0;
				cursor_group=0;
				cursor_x=0;
				cursor_y=0;
				update_cursor_data();
				cursor_ofs=o;
				cursor_yofs=y;
				ofs=o;
				yofs=y;
				clearmsg(msg);
				dirtyview();
				break;
			}
			default:
				if ((msg->data_int==K_Space) && (is_sel_tag)) {
					ht_viewer::handlemsg(msg);
					clearmsg(msg);
					dirtyview();
					return;
				} else if (edit) {
					if (editinput(msg->data_int)) {
						clearmsg(msg);
						dirtyview();
					}
				}
				break;
		}
	} else if (msg->msg==htm_edit_mode) {
		start_edit();
		clearmsg(msg);
	} else if (msg->msg==htm_view_mode) {
		stop_edit();
		clearmsg(msg);
	}
	ht_viewer::handlemsg(msg);
}

int ht_uviewer::editinput(int key)
{
	char *e=get_editable(curline, cursor_e, cursor_group);
	switch (e[1]) {
		case HT_EDIT_BYTE: {
			char *p;
			char *hexchars="0123456789abcdef";
			if ((key<0x80) && (p=strchr(hexchars, key))) {
				int r=p-hexchars;
				int o=tag_get_offset(e);
				int v=0;
				file->seek(o);
				file->read((unsigned char*)&v, 1);

				if (cursor_x) {
					v=(v&0xf0)|r;
				} else {
					v=(v&0x0f)|(r<<4);
				}
				file->seek(o);
				file->write((unsigned char*)&v, 1);
				qdraw();
				cursor_right(1);
				dirtyview();
				return 1;
			}
			break;
		}
		case HT_EDIT_WORD: {
			char *p;
			char *hexchars="0123456789abcdef";
			if ((key<0x80) && (p=strchr(hexchars, key))) {
				int r=p-hexchars;
				int o=tag_get_offset(e);
				int v=0;
				file->seek(o);
				file->read((unsigned char*)&v, 2);

				switch (cursor_x) {
					case 0:
						v=(v&0x0fff)|(r<<12);
						break;
					case 1:
						v=(v&0xf0ff)|(r<<8);
						break;
					case 2:
						v=(v&0xff0f)|(r<<4);
						break;
					case 3:
						v=(v&0xfff0)|r;
						break;
				}
				file->seek(o);
				file->write((unsigned char*)&v, 2);
				qdraw();
				cursor_right(1);
				dirtyview();
				return 1;
			}
			break;
		}
		case HT_EDIT_DWORD: {
			char *p;
			char *hexchars="0123456789abcdef";
			if ((key<0x80) && (p=strchr(hexchars, key))) {
				int r=p-hexchars;
				int o= tag_get_offset(e);
				int v=0;
				file->seek(o);
				file->read((unsigned char*)&v, 4);

				switch (cursor_x) {
					case 0:
						v=(v&0x0fffffff)|(r<<28);
						break;
					case 1:
						v=(v&0xf0ffffff)|(r<<24);
						break;
					case 2:
						v=(v&0xff0fffff)|(r<<20);
						break;
					case 3:
						v=(v&0xfff0ffff)|(r<<16);
						break;
					case 4:
						v=(v&0xffff0fff)|(r<<12);
						break;
					case 5:
						v=(v&0xfffff0ff)|(r<<8);
						break;
					case 6:
						v=(v&0xffffff0f)|(r<<4);
						break;
					case 7:
						v=(v&0xfffffff0)|r;
						break;
				}
				file->seek(o);
				file->write((unsigned char*)&v, 4);
				qdraw();
				cursor_right(1);
				dirtyview();
				return 1;
			}
			break;
		}
		case HT_EDIT_CHAR:
			if ((key>=' ') && (key<0x100)) {
				int o= tag_get_offset(e);
				file->seek(o);
				file->write((unsigned char*)&key, 1);
				qdraw();
				cursor_right(1);
				dirtyview();
				return 1;
			}
			break;
	}
	return 0;
}

void ht_uviewer::insertsub(hts_sub *sub)
{
	if (lastsub) lastsub->next=sub;
	sub->prev=lastsub;
	lastsub=sub;
	if (!firstsub) firstsub=sub;
	if (!currentsub) {
		currentsub=sub;
		currentsub->start_line_address(&ofs, &yofs);
	}
	if (!lastsub) lastsub=sub;
}

void ht_uviewer::complete()
{
	ADDR _ofs=ofs;
	hts_sub *_sub=currentsub;
	int _yofs=yofs;
	int y=0;
	char linebuf[1024];	/* FIXME: possible buffer overflow ! */
	char *str=linebuf;
	cursor_group=0;
	cursor_sub=0;
	cursor_x=0;
	cursor_y=0;
	while (1) {
		if (!_sub->getline(str, _ofs, _yofs)) assert(0);
		int c=count_editables(str);
		if (c) {
			char *e=get_editable(str, 0, cursor_group);
			is_sel_tag=(e[1]==HT_SEL);
			cursor_eofs=tag_get_offset(e);
			cursor_ofs=_ofs;
			cursor_yofs=_yofs;
			cursor_sub=_sub;
			cursor_y=y;
			break;
		}
		y++;
		if (!seek_next_line(&_sub, &_ofs, &_yofs, 1)) ERROR("no editables/selectables in hts_sub(s)");
	}
}

void ht_uviewer::apply_modifier(char *htstring)
{
	while ((htstring=findtag(htstring))) {
		switch (htstring[1]) {
			case HT_EDIT_BYTE: {
				dword o=tag_get_offset(htstring);
				if (file->bytedirty(o)) {
					tag_set_color(htstring, tag_coloridx_edit_modified);
				}
				htstring+=HT_EDIT_BYTE_LEN;
				break;
			}
			case HT_EDIT_WORD: {
				dword o=tag_get_offset(htstring);
				if ((file->bytedirty(o)) || (file->bytedirty(o+1))) {
					 tag_set_color(htstring, tag_coloridx_edit_modified);
				}
				htstring+=HT_EDIT_WORD_LEN;
				break;
			}
			case HT_EDIT_DWORD: {
				dword o=tag_get_offset(htstring);
				if ((file->bytedirty(o)) || (file->bytedirty(o+1)) || (file->bytedirty(o+2)) || (file->bytedirty(o+3))){
					 tag_set_color(htstring, tag_coloridx_edit_modified);
				}
				htstring+=HT_EDIT_DWORD_LEN;
				break;
			}
			case HT_EDIT_CHAR: {
				dword o=tag_get_offset(htstring);
				if (file->bytedirty(o)) {
					 tag_set_color(htstring, tag_coloridx_edit_modified);
				}
				htstring+=HT_EDIT_CHAR_LEN;
				break;
			}
			default:
				htstring+=2;
				break;
		}
	}
}

int ht_uviewer::jump(ADDR tofs, int to_history, ht_view *focused)
{
	char line[1024];	/* FIXME: possible buffer overflow ! */
	char *str=line;
	int y;
	ADDR o;
	hts_sub *s=firstsub;
	while (s) {
		if (s->jump(tofs, &o, &y)) {
			char *e;
			while (1) {
				s->getline(str, o, y);
				e=get_editable(str, 0, cursor_group);
				if (e) break;
				if (!seek_next_line(&s, &o, &y, 1)) return 0;
			}
			int xp=0;
			while (1) {
				if (tofs==tag_get_offset(e)) break;
				xp++;
				e=get_editable(e, 1, 0);
				if (!e) {
					xp=0;
					break;
				}
			}
			if (to_history) {
				void *vs=view_state_store();
				if (vs) {
					vstate *v=new vstate;
					v->focused=focused;
					v->view=this;
					v->data=vs;
					vs_history->push(v);
				}
			}
			strcpy(curline, str);
			currentsub=s;
			cursor_sub=s;
			cursor_e=xp;
			cursor_group=0;
			cursor_x=0;
			cursor_y=0;
			int scroll=1;
			int ny=0;
			for (int i=0; i<size.h; i++) {
				if ((line_sub[i]==s) && (line_ofs[i]==o) && (line_yofs[i]==y)) {
					scroll=0;
					cursor_y=ny;
					break;
				}
				ny++;
			}
			if (scroll) {
				ofs=o;
				yofs=y;
			}
			cursor_ofs=o;
			cursor_yofs=y;
			update_cursor_data();
			dirtyview();
			return 1;
		}
		s=s->next;
	}
	return 0;
}

void ht_uviewer::qdraw()
{
	int cursornotfound=0;
	bounds b;
	getbounds(&b);
	char linebuf[1024];	/* FIXME: possible buffer overflow ! */
	char *str=linebuf;
restart:
	clear(C_BLACK);
	ADDR _ofs=ofs;
	int _yofs=yofs;
	hts_sub *_sub=currentsub;
	int foundcursor=0;
	for (int y=0; y<b.h; y++) {
		if (!_sub->getline(str, _ofs, _yofs)) assert(0);
		if (!foundcursor) {
			if ((cursornotfound) || (y==cursor_y)) {
				int c=count_editables(str);
				char *e=get_editable(str, 0, cursor_group);
				if (_sub==cursor_sub) {
					for (int i=0; i<c; i++) {
						ADDR o=tag_get_offset(e);
						if (o==cursor_eofs) {
found:
							cursor_ofs=_ofs;
							cursor_yofs=_yofs;
							cursor_e=i;
							cursor_y=y;
							strcpy(curline, str);
							update_cursor_data();
							foundcursor=1;
							return;
						}
						e=get_editable(e, 1, 0);
					}
				}
				if ((cursor_y==y) && (!foundcursor)) {
					if (cursornotfound) {
						e=get_editable(str, 1, 0);
						if (!e) assert(0);
						goto found;
					} else {
						cursornotfound=1;
						goto restart;
					}
				}
			}
		}

		if (!seek_next_line(&_sub, &_ofs, &_yofs, 1)) break;
	}
	if (!foundcursor) {
		assert(0);	/* FIXME: debug */
/*		scroll_down(1);
		goto restart;*/
	}
}

void ht_uviewer::draw()
{
	clear(windialog_color_body);
	int cursornotfound=0;
	bounds b;
	getbounds(&b);
	char linebuf[1024];	/* FIXME: possible buffer overflow ! */
	char *str=linebuf;
restart:
	ADDR _ofs=ofs;
	int _yofs=yofs;
	hts_sub *_sub=currentsub;
	int foundcursor=0;
	for (int y=0; y<b.h; y++) {
		if (!_sub->getline(str, _ofs, _yofs)) assert(0);
		line_sub[y]=_sub;
		line_ofs[y]=_ofs;
		line_yofs[y]=_yofs;
		if (edit) apply_modifier(str);
		if (!foundcursor) {
			if ((cursornotfound) || (y==cursor_y)) {
				int c=count_editables(str);
				char *e=get_editable(str, 0, cursor_group);
				if (_sub==cursor_sub) {
					for (int i=0; i<c; i++) {
						ADDR o=tag_get_offset(e);
						if ((o==cursor_eofs) && (is_sel_tag==(e[1]==HT_SEL))) {
found:
							cursor_ofs=_ofs;
							cursor_yofs=_yofs;
							cursor_e=i;
							cursor_y=y;
							strcpy(curline, str);
							update_cursor_data();
							if (tag_is_editable(e)) {
								if (edit) {
									int len=tag_strlen(str);
									int v=len-tag_strlen(e)+cursor_x-xofs*2;
									if (v>=0) setcursor(v, y); else hidecursor();
									tag_set_color(e, tag_coloridx_edit_edited);
								} else {
									tag_set_color(e, tag_coloridx_edit_selected);
								}
							} else {
								hidecursor();
								tag_set_color(e, tag_coloridx_sel_selected);
							}
							foundcursor=1;
							break;
						}
						e=get_editable(e, 1, 0);
					}
				}
				if ((cursor_y==y) && (!foundcursor)) {
					if (cursornotfound) {
						e=get_editable(str, 1, 0);
						if (!e) assert(0);
						goto found;
					} else {
						cursornotfound=1;
						goto restart;
					}
				}
			}
		}

//		lprint(0, y, 7, b.w, str+xofs*2);
		htprint(0, y, b.w, xofs*2, str);
		if (!seek_next_line(&_sub, &_ofs, &_yofs, 1)) break;
	}
//     printf(0, b.h, viewer_color_normal, "%c, mem=%d, x=%d, y=%d, csub=%d, cofs=%08x, cyofs=%d, ceofs=%08x, ce=%d", edit ? 'E' : 'V', get_free_mem(), cursor_x, cursor_y, cursor_sub, cursor_ofs, cursor_yofs, cursor_eofs, cursor_e);
}

int ht_uviewer::htprint(int x, int y, int maxlen, int ofs, char *text)
{
	char *b=text, *r=b;
	palette *pal=pal_grp->get(tag_palette_str);
	int c=tag_color_normal, cx=x, cy=y, n=0, d, lastv=0;
	while ((r=findtag(b)) && (n<maxlen)) {
		if (ofs<0) ofs=0;
		if (r!=b) {
			d=lprint(cx, cy, c, MIN(r-b-ofs, maxlen-n), b+ofs);
			n+=d;
			cx+=d;
			if (r-b>0) ofs-=r-b;
		}
		if (ofs<0) ofs=0;
		if (r[0]=='\e') {
			switch ((unsigned char)r[1]) {
				case HT_EDIT_BYTE: {
					int c=pal[hexb(r+12)].value;
					int v=hexb(r+10);
					lastv=v;
					d=lprint(cx, cy, c, MIN(2-ofs, maxlen-n), r+10+ofs);
					n+=d;
					cx+=d;
					ofs-=2;
					r+=HT_EDIT_BYTE_LEN;
					break;
				}
				case HT_EDIT_WORD: {
					int c=pal[hexb(r+14)].value;
					int v=hexw(r+10);
					lastv=v;
					d=lprint(cx, cy, c, MIN(4-ofs, maxlen-n), r+10+ofs);
					n+=d;
					cx+=d;
					ofs-=4;
					r+=HT_EDIT_WORD_LEN;
					break;
				}
				case HT_EDIT_DWORD: {
					int c=pal[hexb(r+18)].value;
					int v=hexd(r+10);
					lastv=v;
					d=lprint(cx, cy, c, MIN(8-ofs, maxlen-n), r+10+ofs);
					n+=d;
					cx+=d;
					ofs-=8;
					r+=HT_EDIT_DWORD_LEN;
					break;
				}
				case HT_EDIT_TIME: {
//                         int c=pal[hexb(r+18)].value;
					int v=hexd(r+10);
					lastv=v;
/*					tm *time=gmtime((time_t *)&v);
					char line[32];
					sprintf(line, "%02d:%02d:%02d %02d.%02d.%04d", time->tm_hour, time->tm_min, time->tm_sec, time->tm_mday, time->tm_mon+1, time->tm_year+1900);
					d=lprint(cx, cy, c, MIN(19-ofs, maxlen-n), line+ofs);*/
					d=2;
					n+=d;
					cx+=d;
					ofs-=8;
					r+=HT_EDIT_TIME_LEN;
					break;
				}
				case HT_EDIT_CHAR: {
					int c=pal[hexb(r+12)].value;
					int v=hexb(r+10);
					lastv=v;
					char lb[2];
					lb[0]=v;
					if (lb[0]<' ') lb[0]=zerofillchar;
					lb[1]=0;
					d=lprint(cx, cy, c, MIN(1-ofs, maxlen-n), lb+ofs);
					n+=d;
					cx+=d;
					ofs-=1;
					r+=HT_EDIT_CHAR_LEN;
					break;
				}
				case HT_SEL: {
					int c=pal[hexb(r+12)].value;
					int l=hexb(r+10);
					d=lprint(cx, cy, c, MIN(l-ofs, maxlen-n), r+HT_SEL_LEN(0)+ofs);
					n+=d;
					cx+=d;
					ofs-=1;
					r+=HT_SEL_LEN(l);
					break;
				}
				case HT_EXPL: {
					int i=hexd(r+2);
					int_hash *h=find_expl(i);
					char *s;
					if ((h) && (s=matchhash(lastv, h))) {
						int l=strlen(s);
						d=lprint(cx, cy, c, MIN(l-ofs, maxlen-n), s+ofs);
						n+=d;
						cx+=d;
						ofs-=l;
					} else {
						d=lprint(cx, cy, c, MIN(3-ofs, maxlen-n), (char*)"???"+ofs);
						n+=d;
						cx+=d;
						ofs-=3;
					}
					r+=HT_EXPL_LEN;
					break;
				}
				case HT_GROUP:
					r+=HT_GROUP_LEN;
					break;
				default:
					assert(0);
					break;
			}
		}
		b=r;
	}
	if (ofs<0) ofs=0;
	int l=strlen(b);
	if (n<maxlen) n+=lprint(cx, cy, c, MIN(l-ofs, maxlen-n), b+ofs);
	return n;
}

void ht_uviewer::update_cursor_data()
{
	cursor_group=MIN(cursor_group, count_groups(curline)-1);
	cursor_e=MIN(cursor_e, count_editables_in_group(curline, cursor_group)-1);
	char *e=get_editable(curline, cursor_e, cursor_group);
	is_sel_tag=(e[1]==HT_SEL);
	cursor_eofs=tag_get_offset(e);
	if (edit) {
		cursor_x=MIN(cursor_x, tag_get_slen(e)-1);
	} else {
		cursor_x=0;
	}
}

void ht_uviewer::cursor_home()
{
	cursor_e=0;
	cursor_x=0;
	update_cursor_data();
}

void ht_uviewer::cursor_end()
{
	cursor_e=count_editables_in_group(curline, cursor_group)-1;
	cursor_x=tag_get_slen(get_editable(curline, cursor_e, cursor_group))-1;
	update_cursor_data();
}

void ht_uviewer::cursor_next_group()
{
	int g=count_groups(curline);
	cursor_group++;
	if (cursor_group>=g) cursor_group=0;
	update_cursor_data();
}

int ht_uviewer::start_edit()
{
	edit=1;
	dirtyview();
	return 1;
}

int ht_uviewer::stop_edit()
{
	edit=0;
	if (file) {
		file->invd();
		dirtyview();
	}
	if (!edit) hidecursor();
	return 1;
}

int ht_uviewer::search(ht_search *s, hts_sub **sub, ADDR *ofs)
{
	while (*sub<=lastsub) {
		int r=(*sub)->search(s, ofs);
		if (r==htsr_found) {
			jump(*ofs, 1, this);
			return 1;
		}
		(*sub)++;
	}
	infobox("string not found.");
	return 0;
}


int ht_uviewer::seek_prev_line(hts_sub **sub, ADDR *ofs, int *yofs, int n)
{
	int o=n;
	while (n) {
		if (n-=(*sub)->prev_line_address(ofs, yofs, n)) {
			if (!(*sub)->prev) break;
			*sub=(*sub)->prev;
			(*sub)->end_line_address(ofs, yofs);
			n--;
		}
	}
	return o-n;
}

int ht_uviewer::seek_next_line(hts_sub **sub, ADDR *ofs, int *yofs, int n)
{
	int o=n;
	while (n) {
		if (n-=(*sub)->next_line_address(ofs, yofs, n)) {
			if (!(*sub)->next) break;
			*sub=(*sub)->next;
			(*sub)->start_line_address(ofs, yofs);
			n--;
		}
	}
	return o-n;
}

int ht_uviewer::seek_prev_editable_line(hts_sub **sub, ADDR *ofs, int *yofs)
{
	ADDR o=*ofs;
	int y=*yofs, n=0;
	hts_sub *s=*sub;
	do {
		if (!seek_prev_line(&s, &o, &y, 1)) return 0;
		s->getline(curline, o, y);
		n++;
	} while (!count_editables(curline));
	*ofs=o;
	*yofs=y;
	*sub=s;
	return n;
}

int ht_uviewer::seek_next_editable_line(hts_sub **sub, ADDR *ofs, int *yofs)
{
	ADDR o=*ofs;
	int y=*yofs, n=0;
	hts_sub *s=*sub;
	do {
		if (!seek_next_line(&s, &o, &y, 1)) return 0;
		s->getline(curline, o, y);
		n++;
	} while (!count_editables(curline));
	*ofs=o;
	*yofs=y;
	*sub=s;
	return n;
}

int ht_uviewer::count_lines_to_top(hts_sub *sub, ADDR ofs, int yofs)
{
	int c=0;
	while (seek_prev_line(&sub, &ofs, &yofs, 1)) c++;
	return c;
}

int ht_uviewer::count_lines_to_bottom(hts_sub *sub, ADDR ofs, int yofs)
{
	int c=0;
	while (seek_next_line(&sub, &ofs, &yofs, 1)) c++;
	return c;
}

int ht_uviewer::ref()
{
	char *e=get_editable(curline, cursor_e, cursor_group);
	int i;
	if ((i=tag_getselid(e))!=-1) {
		ref(i);
		return 1;
	}
	return 0;
}

void ht_uviewer::ref(dword id)
{
/* FIXME: */
	jump(id, 1, this);
}

void ht_uviewer::scroll_up(int n)
{
	if (!seek_prev_line(&currentsub, &ofs, &yofs, n)) assert(0);
}

void ht_uviewer::scroll_down(int n)
{
	if (!seek_next_line(&currentsub, &ofs, &yofs, n)) assert(0);
}

void ht_uviewer::scroll_left(int n)
{
	if (xofs>0) xofs--;
}

void ht_uviewer::scroll_right(int n)
{
	xofs++;
}

int ht_uviewer::update_edit()
{
	if (file) {
		file->flush();
		dirtyview();
	}
	return 1;
}

struct ht_uviewer_vs {
	hts_sub *currentsub, *cursor_sub;
	ADDR ofs;
	int yofs, xofs;
	ADDR cursor_eofs;
	ADDR cursor_ofs;
	int cursor_yofs;
	int cursor_e, cursor_x, cursor_y, cursor_group;
	int is_sel_tag;
};

void ht_uviewer::view_state_load(void *data)
{
	ht_uviewer_vs *vs=(ht_uviewer_vs *)data;
	currentsub=vs->currentsub;
	cursor_sub=vs->cursor_sub;
	ofs=vs->ofs;
	yofs=vs->yofs;
	xofs=vs->xofs;
	cursor_eofs=vs->cursor_eofs;
	cursor_ofs=vs->cursor_ofs;
	cursor_yofs=vs->cursor_yofs;
	cursor_e=vs->cursor_e;
	cursor_x=vs->cursor_x;
	cursor_y=vs->cursor_y;
	cursor_group=vs->cursor_group;
	is_sel_tag=vs->is_sel_tag;
}

void *ht_uviewer::view_state_store()
{
	ht_uviewer_vs *vs=new ht_uviewer_vs;
	vs->currentsub=currentsub;
	vs->cursor_sub=cursor_sub;
	vs->ofs=ofs;
	vs->yofs=yofs;
	vs->xofs=xofs;
	vs->cursor_eofs=cursor_eofs;
	vs->cursor_ofs=cursor_ofs;
	vs->cursor_yofs=cursor_yofs;
	vs->cursor_e=cursor_e;
	vs->cursor_x=cursor_x;
	vs->cursor_y=cursor_y;
	vs->cursor_group=cursor_group;
	vs->is_sel_tag=is_sel_tag;
	return vs;
}

int ht_uviewer::cursor_up(int n)
{
	int s=0;
	n=seek_prev_line(&cursor_sub, &cursor_ofs, &cursor_yofs, n);
	if (n) {
		int scrolltop=0;
		cursor_sub->getline(curline, cursor_ofs, cursor_yofs);
		if (!count_editables(curline)) {
			int r=seek_prev_editable_line(&cursor_sub, &cursor_ofs, &cursor_yofs);
			if (!r) {
				r=-seek_next_editable_line(&cursor_sub, &cursor_ofs, &cursor_yofs);
				scrolltop=1;
			} else s++;
			n+=r;
		} else s++;
		cursor_y-=n;
		if (cursor_y<0) {
			scroll_up(-cursor_y);
			cursor_y=0;
		}
		if (scrolltop) {
			int c=count_lines_to_top(currentsub, ofs, yofs);
			if (c>0) {
				scroll_up(c);
				cursor_y+=c;
			}
		}
		update_cursor_data();
	}
	return s;
}

int ht_uviewer::cursor_down(int n)
{
	int s=0;
	n=seek_next_line(&cursor_sub, &cursor_ofs, &cursor_yofs, n);
	if (n) {
		bounds b;
		getbounds(&b);
		int scrolltop=0;
		cursor_sub->getline(curline, cursor_ofs, cursor_yofs);
		if (!count_editables(curline)) {
			int r=seek_next_editable_line(&cursor_sub, &cursor_ofs, &cursor_yofs);
			if (!r) {
				r=-seek_prev_editable_line(&cursor_sub, &cursor_ofs, &cursor_yofs)+1;
				scrolltop=1;
			} else s++;
			n+=r;
		} else s++;
		cursor_y+=n;
		if (cursor_y>b.h-1) {
			scroll_down(cursor_y-b.h+1);
			cursor_y=b.h-1;
		}
		if (scrolltop) {
			int c=count_lines_to_bottom(currentsub, ofs, yofs)-b.h+1;
			if (c>0) {
				scroll_down(c);
				cursor_y-=c;
			}
		}
		update_cursor_data();
	}
	return s;
}

int ht_uviewer::cursor_left(int n)
{
	int o=n;
	while (n--) {
		if (cursor_x<=0) {
			if (cursor_e>0) {
				cursor_e--;
				cursor_x=666;
			} else {
				if (cursor_up(1)) cursor_end(); else break;
			}
		} else cursor_x--;
	}
	update_cursor_data();
	return o-n;
}

int ht_uviewer::cursor_right(int n)
{
	int o=n;
	int c=count_editables_in_group(curline, cursor_group);
	while (n--) {
		char *e=get_editable(curline, cursor_e, cursor_group);
		int s=1;
		if (edit) s=tag_get_slen(e);
		if (cursor_x+1>=s) {
			if (cursor_e<c-1) {
				cursor_e++;
				cursor_x=0;
			} else {
				if (cursor_down(1)) cursor_home(); else break;
			}
		} else cursor_x++;
	}
	update_cursor_data();
	return o-n;
}

/*
 *	CLASS ht_logviewer
 */

ht_logviewer::ht_logviewer(bounds *b, int _maxlinecount)
: ht_viewer(b, "log", 0, 0)
{
	maxlinecount=_maxlinecount;
	lines=new ht_string_collection();
	ofs=0;
	xofs=0;
}

ht_logviewer::~ht_logviewer()
{
	delete lines;
}

int ht_logviewer::cursor_up(int n)
{
	ofs-=n;
	if (ofs<0) ofs=0;
	return n;
}

int ht_logviewer::cursor_down(int n)
{
	int c=lines->count();
	if (c>=size.h) {
		ofs+=n;
		if (ofs>c-size.h) ofs=lines->count()-size.h;
	}
	return n;
}

void ht_logviewer::deletefirstline()
{
	lines->del(0);
}

void ht_logviewer::draw()
{
	clear(windialog_color_body);
	int c=lines->count();
	for (int i=0; i<size.h; i++) {
		if (i+ofs>=c) break;
		char *s=lines->get(i+ofs);
		int l=strlen(s);
		if (xofs<l) print(0, i, windialog_color_body, s+xofs);
	}
}

void ht_logviewer::insertline(char *line)
{
	int c=lines->count();
	if (c>=maxlinecount) deletefirstline();
	lines->insert(line);
	if (c>=size.h-1) {
		ofs=c-size.h+1;
	}
	dirtyview();
}

void ht_logviewer::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Right: case K_Control_Right:
				xofs+=2;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Left: case K_Control_Left:
				if (xofs-2>=0) xofs-=2;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Control_PageUp:
				ofs=0;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Control_PageDown:
				ofs=lines->count()-size.h;
				dirtyview();
				clearmsg(msg);
				break;
		}
	}
	ht_viewer::handlemsg(msg);
}

/*
 *	CLASS hts_sub
 */

hts_sub::hts_sub(htstream *_file)
{
	prev=0;
	next=0;
	file=_file;
}

hts_sub::~hts_sub()
{
}

ABSTRACT int hts_sub::getline(char *line, ADDR ofs, int yofs)
{
	abstract();
	return 0;
}

ABSTRACT void hts_sub::start_line_address(ADDR *ofs, int *yofs)
{
	abstract();
}

ABSTRACT void hts_sub::end_line_address(ADDR *ofs, int *yofs)
{
	abstract();
}

ABSTRACT int hts_sub::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	abstract();
	return 0;
}

ABSTRACT int hts_sub::next_line_address(ADDR *ofs, int *yofs, int n)
{
	abstract();
	return 0;
}

STUB int hts_sub::search(ht_search *s, ADDR *ofs)
{
	return htsr_not_supported;
}

STUB int hts_sub::jump(ADDR tofs, ADDR *ofs, int *yofs)
{
	return 0;
}

/*
 *	CLASS hts_linear_sub
 */

hts_linear_sub::hts_linear_sub(htstream *file, int ofs, int size)
: hts_sub(file)
{
	fofs=ofs;
	fsize=size;
}

hts_linear_sub::~hts_linear_sub()
{
}

int hts_linear_sub::search(ht_search *s, ADDR *ofs)
{
	if (s->type==htst_bin) {
		bounds b;
		get_std_progress_indicator_metrics(&b);
		ht_progress_indicator *progress_indicator=new ht_progress_indicator(&b, "searching...%d%% complete");
		int ret=htsr_not_found;
		if (search_str(file, *ofs, fofs+fsize, s->length, s->data, ofs, progress_indicator, s->caseinsensitive, s->idle))
			ret=htsr_found;
		delete progress_indicator;
		return ret;
	}
	return htsr_not_supported;
}

/*
 *	CLASS hts_hex
 */

hts_hex::hts_hex(htstream *file, int ofs, int size)
: hts_linear_sub(file, ofs, size)
{
}

hts_hex::~hts_hex()
{
}

int hts_hex::getline(char *line, ADDR ofs, int yofs)
{
	unsigned char buf[16];
	file->seek(ofs);
	int c=file->read(buf, 16);
	c=MIN(c, (int)(fofs+fsize-ofs));
	if (c<=0) return 0;
	char *l=line;
	l=mkhexd(l, ofs);
	*l++=' ';
	for (int i=0; i<16; i++) {
		if (i<c) {
			l=mktag_edit_byte(l, ofs+i, buf[i], tag_coloridx_edit_unselected);
		} else {
			*l++=' ';
			*l++=' ';
		}
		*l++=' ';
	}
	l=mktag_group(l);
	for (int i=0; i<16; i++) {
		if (i<c) {
			l=mktag_edit_char(l, ofs+i, buf[i], tag_coloridx_edit_unselected);
		} else {
			*l++=' ';
		}
	}
	*l=0;
	return 1;
}

void hts_hex::start_line_address(ADDR *ofs, int *yofs)
{
	(*ofs)=fofs;
	(*yofs)=0;
}

void hts_hex::end_line_address(ADDR *ofs, int *yofs)
{
	(*ofs)=((fsize-1) & ~0xf)+fofs;
	(*yofs)=0;
}

int hts_hex::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	int c=0;
	while (n--) {
		if (*ofs-16>*ofs) break;
		if (*ofs-16<fofs) break;
		(*ofs)-=16;
		c++;
/*        if (*ofs-16<fofs) break;
		(*ofs)-=16;
		c++;*/
	}
	return c;
}

int hts_hex::next_line_address(ADDR *ofs, int *yofs, int n)
{
	int c=0;
	while (n--) {
		if (*ofs+16>=fofs+fsize) break;
		(*ofs)+=16;
		c++;
	}
	return c;
}

int hts_hex::jump(ADDR tofs, ADDR *ofs, int *yofs)
{
	if (((tofs&~15)>=fofs) && ((tofs&~15)<fofs+fsize)) {
		*ofs=tofs&~15;
		*yofs=0;
		return 1;
	}
	return 0;
}

/*
 *	CLASS hts_unasm
 */

hts_unasm::hts_unasm(htstream *file, ADDR ofs, int size, int use32)
: hts_linear_sub(file, ofs, size)
{
	if (use32) {
		unasm=new x86dis(X86_OPSIZE32, X86_ADDRSIZE32);
	} else {
		unasm=new x86dis(X86_OPSIZE16, X86_ADDRSIZE16);
	}
}

hts_unasm::~hts_unasm()
{
	delete unasm;
}

int hts_unasm::getline(char *line, ADDR ofs, int yofs)
{
	unsigned char buf[15];
	file->seek(ofs);
	int c=file->read(buf, 15);
	c=MIN(c, (int)(fofs+fsize-ofs));
	if (c<=0) return 0;
	CPU_ADDR caddr;
	caddr.addr32=0;
	x86dis_insn *insn=(x86dis_insn*)unasm->decode(buf, c, caddr);
	char *s=unasm->str(insn, X86DIS_STYLE_OPTIMIZE_ADDR | X86DIS_STYLE_EXPLICIT_MEMSIZE);
	char *l=line;
	c=MIN(c, unasm->getsize(insn));
	l=mkhexd(l, ofs);
	*l++=' ';
	for (int i=0; i<15; i++) {
		if (i<c) {
			l=mktag_edit_byte(l, ofs+i, buf[i], tag_coloridx_edit_unselected);
		} else {
			*l++=' ';
			*l++=' ';
		}
//		*l++=' ';
	}
	*l++=' ';
	strcpy(l, s);
//     l+=sprintf(l, " %s", s);
	return 1;
}

void hts_unasm::start_line_address(ADDR *ofs, int *yofs)
{
	(*ofs)=fofs;
	(*yofs)=0;
}

void hts_unasm::end_line_address(ADDR *ofs, int *yofs)
{
	(*ofs)=fofs+fsize;
	(*yofs)=0;
	prev_line_address(ofs, yofs, 1);
}

int hts_unasm::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	unsigned char buf[15*50], *bufp=buf;
	int offsets[15*50];
	int *of=offsets;
	int r=n<6 ? 6*15 : n*15;
	ADDR o=*ofs-r;
	int c=r, d;
	int s;
	if (o<fofs) {
		c-=fofs-o;
		o=fofs;
	}
	if (o+c>fofs+fsize) {
		c=fofs+fsize-o;
	}
	if (c<=0) return 0;
	file->seek(o);
	d=file->read(buf, c);
	CPU_ADDR caddr;
	caddr.addr32=0;
	do {
		x86dis_insn *insn=(x86dis_insn *)unasm->decode(bufp, d, caddr);
		s=insn->size;
//     	s=unasm->getsize(bufp, d);
		d-=s;
		*(of++)=o;
		o+=s;
		bufp+=s;
	} while (o<=*ofs);
	if (of-n-1<offsets) {
		*ofs=*(offsets);
		return of-offsets-1;
	} else {
		*ofs=*(of-n-1);
		return n;
	}
}

int hts_unasm::next_line_address(ADDR *ofs, int *yofs, int n)
{
	unsigned char buf[15];
	int c=0, s, z;
	CPU_ADDR caddr;
	caddr.addr32=0;
	while (n--) {
		z=MIN(15, fofs+fsize-*ofs);
		file->seek(*ofs);
		z=file->read(buf, z);
		if (z<=0) break;
		x86dis_insn *insn=(x86dis_insn *)unasm->decode(buf, z, caddr);
		s=insn->size;
//          s=unasm->getsize(buf, z);
		if (*ofs+s>=fofs+fsize) break;
		*ofs+=s;
		c++;
	}
	return c;
}

int hts_unasm::jump(ADDR tofs, ADDR *ofs, int *yofs)
{
	if (((tofs&~15)>=fofs) && ((tofs&~15)<fofs+fsize)) {
		*ofs=tofs&~15;
		*yofs=0;
		return 1;
	}
	return 0;
}

/*
 *	CLASS hts_text
 */

hts_text::hts_text(htstream *file, const char *_text) : hts_sub(file)
{
	text=(char*)malloc(strlen(_text));
	strcpy(text, _text);
}

hts_text::~hts_text()
{
	if (text) free(text);
}

int hts_text::getline(char *line, ADDR ofs, int yofs)
{
	strcpy(line, text);
	return 1;
}

void hts_text::start_line_address(ADDR *ofs, int *yofs)
{
	*ofs=0;
	*yofs=0;
}

void hts_text::end_line_address(ADDR *ofs, int *yofs)
{
	*ofs=0;
	*yofs=0;
}

int hts_text::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	return 0;
}

int hts_text::next_line_address(ADDR *ofs, int *yofs, int n)
{
	return 0;
}

/*
 *	CLASS hts_mask
 */

hts_mask::hts_mask(htstream *file, int _start) : hts_sub(file)
{
	start=_start;
}

hts_mask::~hts_mask()
{
}

int hts_mask::getline(char *line, ADDR ofs, int yofs)
{
	char *l=temp;
	hts_mask_data *d=(hts_mask_data*)masks.get(ofs);
	tag_strcpy(temp, d->tagstring);
//	strcpy((char*)temp, masks.get(ofs));
	int c=count_editables(l);
	int g=count_groups(l);
	for (int j=0; j<g; j++) {
		for (int i=0; i<c; i++) {
			l=get_editable((char*)&temp, i, j);
			if (tag_is_editable(l)) {
				int o=tag_get_offset(l)+start;
				tag_set_offset(l, o);
				int v=0;
				file->seek(o);
				file->read(&v, tag_get_size(l));
				tag_set_value(l, v);
			}
		}
	}
	strcpy(line, temp);
	return 1;
}

void hts_mask::start_line_address(ADDR *ofs, int *yofs)
{
	*ofs=0;
	*yofs=0;
}

void hts_mask::end_line_address(ADDR *ofs, int *yofs)
{
	*ofs=masks.count()-1;
	*yofs=0;
}

int hts_mask::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	int r;
	if (*ofs<(dword)n) {
		r=*ofs;
		*ofs=0;
	} else {
		r=n;
		*ofs-=n;
	}
	return r;
}

int hts_mask::next_line_address(ADDR *ofs, int *yofs, int n)
{
	int r=n;
	*ofs+=n;
	if ((int)*ofs>masks.count()-1) {
		r-=*ofs-masks.count()+1;
		*ofs=masks.count()-1;
	}
	return r;
}

void hts_mask::add_mask(char *tagstring)
{
	hts_mask_data *d=new hts_mask_data();
	d->tagstring=tag_strdup(tagstring);
	masks.insert(d);
}

void hts_mask::add_mask_table(char **tagstring)
{
	while (*tagstring) add_mask(*(tagstring++));
}

/*
 *	INIT
 */

int init_obj()
{
	palette *tag_palette=pal_grp->create(tag_palette_str, tag_palette_colors);
	pal_grp->set_entry(tag_palette, tag_coloridx_normal, "normal", COLOR(C_WHITE, C_BLACK));
	pal_grp->set_entry(tag_palette, tag_coloridx_edit_selected, "editable: selected", COLOR(C_GREEN, C_BLACK));
	pal_grp->set_entry(tag_palette, tag_coloridx_edit_unselected, "editable: unselected", COLOR(C_LIGHT(C_BLACK), C_BLACK));
	pal_grp->set_entry(tag_palette, tag_coloridx_edit_edited, "editable: edited", COLOR(C_RED, C_BLACK));
	pal_grp->set_entry(tag_palette, tag_coloridx_edit_modified, "editable: modified", COLOR(C_CYAN, C_BLACK));
	pal_grp->set_entry(tag_palette, tag_coloridx_sel_selected, "selectable: selected", COLOR(C_LIGHT(C_WHITE), C_BLUE));
	pal_grp->set_entry(tag_palette, tag_coloridx_sel_unselected, "selectable: unselected", COLOR(C_CYAN, C_BLACK));

	return 1;
}

/*
 *	DONE
 */

int done_obj()
{
	return 1;
}

